@@ -2,6 +2,8 @@
 use num_bigint::BigInt;
 #[cfg(feature = "num-bigint")]
 use num_traits::cast::ToPrimitive;
+#[cfg(feature = "python")]
+use pyo3::{IntoPyObject, IntoPyObjectRef};
 
 use std::ops::Range;
 
@@ -17,6 +19,7 @@ pub trait AbstractNumberDecoder {
 
 /// A number that can be either an [i64] or a [BigInt](num_bigint::BigInt)
 #[derive(Debug, Clone, PartialEq)]
+#[cfg_attr(feature = "python", derive(IntoPyObject, IntoPyObjectRef))]
 pub enum NumberInt {
     Int(i64),
     #[cfg(feature = "num-bigint")]
@@ -80,7 +83,7 @@ impl AbstractNumberDecoder for NumberFloat {
         if !positive {
             // we started with a minus sign, so the first digit is at index + 1
             index += 1;
-        };
+        }
         let first2 = if positive { Some(&first) } else { data.get(index) };
 
         if let Some(digit) = first2 {
@@ -113,23 +116,12 @@ impl AbstractNumberDecoder for NumberFloat {
 
 /// A number that can be either a [NumberInt] or an [f64]
 #[derive(Debug, Clone, PartialEq)]
+#[cfg_attr(feature = "python", derive(IntoPyObject, IntoPyObjectRef))]
 pub enum NumberAny {
     Int(NumberInt),
     Float(f64),
 }
 
-#[cfg(feature = "python")]
-impl pyo3::ToPyObject for NumberAny {
-    fn to_object(&self, py: pyo3::Python<'_>) -> pyo3::PyObject {
-        match self {
-            Self::Int(NumberInt::Int(int)) => int.to_object(py),
-            #[cfg(feature = "num-bigint")]
-            Self::Int(NumberInt::BigInt(big_int)) => big_int.to_object(py),
-            Self::Float(float) => float.to_object(py),
-        }
-    }
-}
-
 impl From<NumberAny> for f64 {
     fn from(num: NumberAny) -> Self {
         match num {
@@ -205,7 +197,7 @@ impl IntParse {
         if !positive {
             // we started with a minus sign, so the first digit is at index + 1
             index += 1;
-        };
+        }
         let first2 = if positive { Some(&first) } else { data.get(index) };
         let first_value = match first2 {
             Some(b'0') => {
@@ -420,7 +412,7 @@ impl AbstractNumberDecoder for NumberRange {
         if !positive {
             // we started with a minus sign, so the first digit is at index + 1
             index += 1;
-        };
+        }
 
         match data.get(index) {
             Some(b'0') => {
@@ -448,7 +440,7 @@ impl AbstractNumberDecoder for NumberRange {
             Some(digit) if (b'1'..=b'9').contains(digit) => (),
             Some(_) => return json_err!(InvalidNumber, index),
             None => return json_err!(EofWhileParsingValue, index),
-        };
+        }
 
         index += 1;
         for _ in 0..18 {
@@ -506,13 +498,13 @@ fn consume_exponential(data: &[u8], mut index: usize) -> JsonResult<usize> {
         Some(v) if v.is_ascii_digit() => (),
         Some(_) => return json_err!(InvalidNumber, index),
         None => return json_err!(EofWhileParsingValue, index),
-    };
+    }
 
     match data.get(index) {
         Some(v) if v.is_ascii_digit() => (),
         Some(_) => return json_err!(InvalidNumber, index),
         None => return json_err!(EofWhileParsingValue, index),
-    };
+    }
     index += 1;
 
     while let Some(next) = data.get(index) {
@@ -531,7 +523,7 @@ fn consume_decimal(data: &[u8], mut index: usize) -> JsonResult<usize> {
         Some(v) if v.is_ascii_digit() => (),
         Some(_) => return json_err!(InvalidNumber, index),
         None => return json_err!(EofWhileParsingValue, index),
-    };
+    }
     index += 1;
 
     while let Some(next) = data.get(index) {
